from typing import Any, Coroutine, TypeAlias, TypeVar, TYPE_CHECKING, Union
from pydantic import BaseModel

if TYPE_CHECKING:  # pragma: no cover
    from yc_lockbox._models import Operation, YandexCloudError

T = TypeVar("T", bound=BaseModel)

YandexCloudGenericResponse: TypeAlias = (
    Coroutine[Any, Any, Union["Operation", "YandexCloudError"]] | Union["Operation", "YandexCloudError"]
)  # pragma: no cover



import jwt
import time
import json
import enum
import yc_lockbox._constants as C

from abc import ABC, abstractmethod
from functools import lru_cache
from typing import Any, TypeVar
from yc_lockbox._exceptions import InvalidCredentials

T = TypeVar("T")


@enum.unique
class CredentialsType(str, enum.Enum):
    IAM_TOKEN = "IAM_TOKEN"  # nosec B105
    OAUTH_TOKEN = "OAUTH_TOKEN"  # nosec B105
    SERVICE_ACCOUNT = "SERVICE_ACCOUNT"  # nosec B105

    @staticmethod
    def allowed_types(stringify: bool = False) -> list | str:
        """
        Returns allowed credentials types.

        :param stringify: Return as string instead list.
        """
        types_ = [name for name, _ in CredentialsType.__members__.items()]

        if stringify:
            return ", ".join(types_)
        return types_


class AbstractHTTPAdapter(ABC):
    """An abstract adapter for make HTTP requests to Yandex Cloud."""

    @staticmethod
    @abstractmethod
    def request(
        method: str,
        url: str,
        data: str | bytes | None = None,
        json: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        response_model: T | None = None,
        raise_for_status: bool = True,
        **kwargs,
    ) -> Any:
        raise NotImplementedError  # pragma: no cover


class AbstractYandexLockboxClient(ABC):
    """An abstract class for operate with Yandex Lockbox service."""

    def __init__(
        self,
        lockbox_base_url: str | None = None,
        payload_lockbox_base_url: str | None = None,
        **kwargs,
    ) -> None:
        self.lockbox_base_url = lockbox_base_url or C.YC_LOCKBOX_BASE_URL
        self.payload_lockbox_base_url = payload_lockbox_base_url or C.YC_LOCKBOX_PAYLOAD_BASE_URL

    @abstractmethod
    def activate_secret(self, *args, **kwargs) -> T | Any:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def add_secret_version(self, *args, **kwargs) -> T | Any:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def create_secret(self, *args, **kwargs) -> T:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def cancel_secret_version_destruction(self, *args, **kwargs) -> T | Any:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def deactivate_secret(self, *args, **kwargs) -> T | Any:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def delete_secret(self, *args, **kwargs) -> Any:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def get_secret(self, secret_id: str, *args, **kwargs) -> T:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def get_secret_payload(self, *args, **kwargs) -> T:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def list_secrets(self, *args, **kwargs) -> list[T]:
        raise NotImplementedError

    @abstractmethod
    def list_secret_access_bindings(self, *args, **kwargs) -> T | Any:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def list_secret_operations(self, *args, **kwargs) -> list[T] | list[Any]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def list_secret_versions(self, *args, **kwargs) -> list[T] | list[Any]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def schedule_secret_version_destruction(self, *args, **kwargs) -> T | Any:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def set_secret_access_bindings(self, *args, **kwargs) -> T | Any:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def update_secret(self, *args, **kwargs) -> T:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def update_secret_access_bindings(self, *args, **kwargs) -> T | Any:
        raise NotImplementedError  # pragma: no cover


class AbstractYandexAuthClient(ABC):
    """An abstract class for authenticate requests for the Yandex.Cloud."""

    def __init__(
        self,
        credentials: str | dict[str, str],
        *,
        auth_base_url: str | None = None,
        **kwargs,
    ) -> None:
        self.credentials = credentials
        self.auth_base_url = auth_base_url or C.YC_IAM_BASE_URL

        self.iam_token: str | None = None
        self._credentials_type: CredentialsType = self._detect_credentials_type()

    @property
    @abstractmethod
    def adapter(self) -> AbstractHTTPAdapter:
        """An abstract property that returns sync or async adapter for HTTP requests."""
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def get_iam_token(self) -> str:
        """An abstract method for get IAM token from Yandex.Cloud."""
        raise NotImplementedError  # pragma: no cover

    def get_auth_headers(self) -> dict[str, str]:
        """Returns auth headers as dict."""
        token = self.get_iam_token()
        return {"Authorization": f"Bearer {token}"}

    @staticmethod
    def is_sa_key(credentials: Any) -> bool:
        """
        A predicate that tells whether the credentials is a service account key.

        :param credentials: Credentials to verify.
        """
        mandatory_attributes = (
            "id",
            "service_account_id",
            "created_at",
            "key_algorithm",
            "public_key",
            "private_key",
        )
        if isinstance(credentials, str):
            try:
                credentials = json.loads(credentials)
            except json.JSONDecodeError:
                credentials = {}

        if isinstance(credentials, dict):
            return all([credentials.get(field) for field in mandatory_attributes])

        return False

    @staticmethod
    def is_iam_token(credentials: Any) -> bool:
        """
        A predicate that tells whether the credentials is a IAM token.

        :param credentials: Credentials to verify.
        """
        if not isinstance(credentials, str):
            return False

        return C.IAM_TOKEN_REGEX.match(credentials)

    @staticmethod
    def is_oauth_token(credentials: Any) -> bool:
        """
        A predicate that tells whether the credentials is a OAuth token.

        :param credentials: Credentials to verify.
        """
        if not isinstance(credentials, str):
            return False

        return C.OAUTH_TOKEN_REGEX.match(credentials) or C.LEGACY_OAUTH_TOKEN_REGEX.match(credentials)

    def _detect_credentials_type(self) -> CredentialsType:
        """Detect and verify credentials type."""
        if self.is_iam_token(self.credentials):
            return CredentialsType.IAM_TOKEN
        elif self.is_sa_key(self.credentials):
            return CredentialsType.SERVICE_ACCOUNT
        elif self.is_oauth_token(self.credentials):
            return CredentialsType.OAUTH_TOKEN

        raise InvalidCredentials(
            f"Invalid credentials. Allowed types: {CredentialsType.allowed_types(stringify=True)}"
        )

    @lru_cache(maxsize=1)
    def _generate_jwt_for_sa(self, aud_path: str = "/iam/v1/tokens") -> str:
        """
        Prepare JWT for request IAM token from service account key.

        :param aud_path: URL path for JWT payload ``aud``.
        """

        if self._credentials_type != CredentialsType.SERVICE_ACCOUNT or not isinstance(self.credentials, dict):
            raise InvalidCredentials(f"Can't generate JWT, invalid credentials type: {self._credentials_type}")

        now = int(time.time())
        sa_id = self.credentials.get("service_account_id")
        key_id = self.credentials.get("id")
        private_key = self.credentials.get("private_key")

        payload = {
            "iss": sa_id,
            "aud": self.auth_base_url + aud_path,
            "iat": now,
            "exp": now + 360,
        }

        headers = {
            "typ": "JWT",
            "alg": C.JWT_ALGORITHM,
            "kid": key_id,
        }

        return jwt.encode(payload, private_key, algorithm=C.JWT_ALGORITHM, headers=headers)



import re
import enum

YC_LOCKBOX_BASE_URL = "https://lockbox.api.cloud.yandex.net/lockbox/v1"
YC_LOCKBOX_PAYLOAD_BASE_URL = "https://payload.lockbox.api.cloud.yandex.net/lockbox/v1"
YC_IAM_BASE_URL = "https://iam.api.cloud.yandex.net"
JWT_ALGORITHM = "PS256"

# regex patterns

IAM_TOKEN_REGEX = re.compile(r"t1\.[A-Z0-9a-z_-]+[=]{0,2}\.[A-Z0-9a-z_-]{86}[=]{0,2}")
OAUTH_TOKEN_REGEX = re.compile(r"y[0-3]_[a-zA-Z0-9_-]+")
LEGACY_OAUTH_TOKEN_REGEX = re.compile(r"AQ[a-zA-Z0-9_-]+")


class RpcError(int, enum.Enum):
    """This class convert grpc digit codes to messages."""

    CANCELLED = 1
    UNKNOWN = 2
    INVALID_ARGUMENT = 3
    DEADLINE_EXCEEDED = 4
    NOT_FOUND = 5
    ALREADY_EXISTS = 6
    PERMISSION_DENIED = 7
    RESOURCE_EXHAUSTED = 8
    FAILED_PRECONDITION = 9
    ABORTED = 10
    OUT_OF_RANGE = 11
    NOT_IMPLEMENTED = 12
    INTERNAL = 13
    UNAVAILABLE = 14
    DATA_LOSS = 15
    UNAUTHENTICATED = 16



import logging
from typing import Type, Optional, Callable, Iterator

from yc_lockbox._abc import AbstractYandexAuthClient, AbstractYandexLockboxClient, AbstractHTTPAdapter
from yc_lockbox._adapters import HTTPAdapter
from yc_lockbox._auth import YandexAuthClient
from yc_lockbox._models import (
    Secret,
    SecretPayload,
    SecretVersion,
    SecretsList,
    SecretVersionsList,
    Operation,
    YandexCloudError,
    INewSecret,
    IUpdateSecret,
    INewSecretVersion,
    BasePaginatedResponse,
)
from yc_lockbox._types import T

logger = logging.getLogger(__name__)


class YandexLockboxClient(AbstractYandexLockboxClient):
    """
    Yandex Lockbox secrets vault client.

    :param credentials: Credentials for authenticate requests.
        Allowed types: service account key, OAuth token, IAM token.
    :param auth_client: Optional client implementation for authenticate requests.
        Defaults to ``YandexAuthClient``.
    :param adapter: HTTP adapter for communicate with Yandex Cloud API.
    :param lockbox_base_url: Lockbox base URL without resource path.
    :param payload_lockbox_base_url: Lockbox payload base URL without resource path.
    :param auth_base_url: IAM base URL without resource path.

    .. note::

        All the values of the secrets are masked, i.e. looks like ``***********``.
        To get the real value of the secret, you need to call the injected methods
        :func:`reveal_text_value()` or :func:`reveal_binary_value()`.

    Usage::

        from yc_lockbox import YandexLockboxClient, Secret

        lockbox = YandexLockboxClient("y0_AgAEXXXXXXXXXXXXXXXXXXXXXXXXX")  # OAuth or IAM token

        secret: Secret = lockbox.get_secret("e6xxxxxxxxxxxxxxxx")
        print(secret.name, secret.status, secret.description)

        payload = secret.payload()

        try:
            value = payload["mykey"]
            print(value.reveal_text_value())
        except KeyError:
            print("Invalid key!")

        print(payload.get("foo"))  # None if not exists without raising exception
        entry = payload[0]  # similar to payload.entries[0]

    Authenticate via service account key::

        import json

        # generate json key for your SA
        # yc iam key create --service-account-name my-sa --output key.json

        with open("./key.json", "r") as infile:
            credentials = json.load(infile)

        lockbox = YandexLockboxClient(credentials)

    """

    def __init__(
        self,
        credentials,
        *,
        auth_client: Optional[Type[AbstractYandexAuthClient]] = YandexAuthClient,
        adapter: Optional[Type[AbstractHTTPAdapter]] = HTTPAdapter,
        lockbox_base_url: str | None = None,
        payload_lockbox_base_url: str | None = None,
    ) -> None:
        super().__init__(
            lockbox_base_url=lockbox_base_url,
            payload_lockbox_base_url=payload_lockbox_base_url,
        )
        self.auth: AbstractYandexAuthClient = auth_client(credentials)
        self.adapter: AbstractHTTPAdapter = adapter

    @property
    def auth_headers(self) -> dict[str, str]:
        """Returns headers for authenticate."""
        return self.auth.get_auth_headers()

    def _inject_client_to_items(self, items: list[T]) -> list[T]:
        """Inject this client to each response model."""
        return list(map(lambda item: item.inject_client(self), items))

    def _seekable_response(
        self, func: Callable[..., BasePaginatedResponse], entrypoint: str, *args, **kwargs
    ) -> Iterator[T]:
        """
        Requests all data from the API using the generators, instead of a page-by-page response.
        This method does not works with dict. Be careful.
        Returns list of objects from model entrypoint.

        :param func: Adapter func to get data from API.
        :param entrypoint: Response attribute that contains list of useful data items.
        :param args: Arguments for func. Will be passed when called inside.
        :param kwargs: Keyword arguments for func. Similar to ``args``.
        """

        if kwargs.get("params") is None:
            raise TypeError(
                "This method works only with query string parameters. Check keyword arguments to resolve it."
            )

        next_token = ""  # nosec B105

        while next_token is not None:
            # There could potentially be a problem if some request doesn't have a 'pageToken' in query string params.
            # However, this is already a question for a non-consistent API, I think.
            # An unlikely story, but worth keeping in mind.
            kwargs["params"]["pageToken"] = next_token

            response = func(*args, **kwargs)
            next_token = response.next_page_token  # None or a new token from the API

            if not hasattr(response, entrypoint):
                raise AttributeError(f"Entrypoint {entrypoint} not exists in response model.")

            for item in getattr(response, entrypoint):
                if not hasattr(item, "inject_client"):
                    raise AttributeError(
                        f"Incorrect item. Method 'inject_client' is not exists in {item.__class__} {type(item)}"
                    )
                item.inject_client(self)
                yield item

    def activate_secret(self, secret_id: str, raise_for_status: bool = True) -> Operation | YandexCloudError:
        """
        Activates the specified secret.

        :param secret_id: Secret indentifier.
        :param raise_for_status: If set to ``False`` returns :class:`YandexCloudError` instead throw exception.
            Defaults to ``True``.
        """
        url = f"{self.lockbox_base_url}/secrets/{secret_id}:activate"
        response = self.adapter.request(
            "POST",
            url,
            headers=self.auth_headers,
            response_model=Operation,
            raise_for_status=raise_for_status,
        )
        response.inject_client(self)
        return response

    def add_secret_version(
        self, secret_id: str, version: INewSecretVersion, raise_for_status: bool = True
    ) -> Operation | YandexCloudError:
        """
        Adds new version based on a previous one.

        :param secret_id: Secret indentifier.
        :param version: A new version object.
        :param raise_for_status: If set to ``False`` returns :class:`YandexCloudError` instead throw exception.
            Defaults to ``True``.
        """
        url = f"{self.lockbox_base_url}/secrets/{secret_id}:addVersion"
        payload = version.model_dump_json(by_alias=True, exclude_none=True)
        response = self.adapter.request(
            "POST",
            url,
            headers=self.auth_headers,
            data=payload,
            response_model=Operation,
            raise_for_status=raise_for_status,
        )
        response.inject_client(self)
        return response

    def create_secret(self, secret: INewSecret, raise_for_status: bool = True) -> Operation | YandexCloudError:
        """
        Creates a secret in the specified folder.

        :param secret: A new secret object.
        :param raise_for_status: If set to ``False`` returns :class:`YandexCloudError` instead throw exception.
            Defaults to ``True``.
        """
        url = f"{self.lockbox_base_url}/secrets"
        payload = secret.model_dump_json(by_alias=True, exclude_none=True)
        response = self.adapter.request(
            "POST",
            url,
            headers=self.auth_headers,
            data=payload,
            response_model=Operation,
            raise_for_status=raise_for_status,
        )
        response.inject_client(self)
        return response

    def cancel_secret_version_destruction(
        self, secret_id: str, version_id: str, raise_for_status: bool = True
    ) -> Operation | YandexCloudError:
        """
        Cancels previously scheduled version destruction, if the version hasn't been destroyed yet.

        :param secret_id: Secret indentifier.
        :param version_id: Secret version id to cancel destruction.
        :param raise_for_status: If set to ``False`` returns :class:`YandexCloudError` instead throw exception.
            Defaults to ``True``.
        """
        url = f"{self.lockbox_base_url}/secrets/{secret_id}:cancelVersionDestruction"
        payload = {"versionId": version_id}
        response = self.adapter.request(
            "POST",
            url,
            headers=self.auth_headers,
            json=payload,
            response_model=Operation,
            raise_for_status=raise_for_status,
        )
        response.inject_client(self)
        return response

    def deactivate_secret(self, secret_id: str, raise_for_status: bool = True) -> Operation | YandexCloudError:
        """
        Deactivate a secret.

        :param secret_id: Secret indentifier.
        :param raise_for_status: If set to ``False`` returns :class:`YandexCloudError` instead throw exception.
            Defaults to ``True``.
        """
        url = f"{self.lockbox_base_url}/secrets/{secret_id}:deactivate"
        response = self.adapter.request(
            "POST",
            url,
            headers=self.auth_headers,
            response_model=Operation,
            raise_for_status=raise_for_status,
        )
        response.inject_client(self)
        return response

    def delete_secret(self, secret_id: str, raise_for_status: bool = True) -> Operation | YandexCloudError:
        """
        Deletes the specified secret.

        :param secret_id: Secret indentifier.
        :param raise_for_status: If set to ``False`` returns :class:`YandexCloudError` instead throw exception.
            Defaults to ``True``.
        """
        url = f"{self.lockbox_base_url}/secrets/{secret_id}"
        response = self.adapter.request(
            "DELETE",
            url,
            headers=self.auth_headers,
            response_model=Operation,
            raise_for_status=raise_for_status,
        )
        response.inject_client(self)
        return response

    def get_secret(self, secret_id: str, raise_for_status: bool = True) -> Secret | YandexCloudError:
        """
        Get lockbox secret by ID.

        :param secret_id: Secret identifier.
        :param raise_for_status: If set to ``False`` returns :class:`YandexCloudError` instead throw exception.
            Defaults to ``True``.
        """
        url = f"{self.lockbox_base_url}/secrets/{secret_id}"
        response = self.adapter.request(
            "GET",
            url,
            headers=self.auth_headers,
            response_model=Secret,
            raise_for_status=raise_for_status,
        )
        response.inject_client(self)
        return response

    def get_secret_payload(
        self,
        secret_id: str,
        version_id: str | None = None,
        raise_for_status: bool = True,
    ) -> SecretPayload | YandexCloudError:
        """
        Get lockbox secret payload by ID and optional version.

        :param secret_id: Secret identifier.
        :param version_id: Secret version. Optional.
        :param raise_for_status: If set to ``False`` returns :class:`YandexCloudError` instead throw exception.
            Defaults to ``True``.
        """
        url = f"{self.payload_lockbox_base_url}/secrets/{secret_id}/payload"
        params = {"version_id": version_id} if version_id else None
        return self.adapter.request(
            "GET",
            url,
            headers=self.auth_headers,
            response_model=SecretPayload,
            raise_for_status=raise_for_status,
            params=params,
        )

    def list_secrets(
        self,
        folder_id: str,
        page_size: int = 100,
        page_token: str | None = None,
        raise_for_status: bool = True,
        iterator: bool = False,
    ) -> SecretsList | Iterator[Secret] | YandexCloudError:
        """
        Retrieves the list of secrets in the specified folder.

        :param folder_id: ID of the folder to list secrets in.
        :param page_size: The maximum number of results per page to return.
            If the number of available results is larger than ``page_size``,
            the service returns a ``next_page_token`` that can be used to get
            the next page of results in subsequent list requests.
            Default value: ``100``.
            The maximum value is ``1000``.
        :param page_token: Page token. To get the next page of results, set ``page_token``
            to the ``next_page_token`` returned by a previous list request.
        :param iterator: Returns all data as iterator (generator) instead paginated result.
        """
        args = (
            "GET",
            f"{self.lockbox_base_url}/secrets",
        )
        kwargs = {
            "headers": self.auth_headers,
            "params": {"folderId": folder_id, "pageSize": page_size, "pageToken": page_token},
            "response_model": SecretsList,
            "raise_for_status": raise_for_status,
        }

        if iterator:
            return self._seekable_response(self.adapter.request, "secrets", *args, **kwargs)

        response = self.adapter.request(*args, **kwargs)
        self._inject_client_to_items(response.secrets)
        return response

    # TODO: implement
    def list_secret_access_bindings(self, *args, **kwargs):
        """Not ready yet."""
        raise NotImplementedError

    # TODO: implement
    def list_secret_operations(self, *args, **kwargs):
        """Not ready yet."""
        raise NotImplementedError

    def list_secret_versions(
        self,
        secret_id: str,
        page_size: int = 100,
        page_token: str | None = None,
        raise_for_status: bool = True,
        iterator: bool = False,
    ) -> SecretVersionsList | Iterator[SecretVersion] | YandexCloudError:
        """
        Retrieves the list of versions of the specified secret.

        :param secret_id: Secret identifier.
        :param page_size: The maximum number of results per page to return.
            If the number of available results is larger than ``page_size``,
            the service returns a ``next_page_token`` that can be used to get
            the next page of results in subsequent list requests.
            Default value: ``100``.
            The maximum value is ``1000``.
        :param page_token: Page token. To get the next page of results, set ``page_token``
            to the ``next_page_token`` returned by a previous list request.
        :param iterator: Returns all data as iterator (generator) instead paginated result.
        """
        args = (
            "GET",
            f"{self.lockbox_base_url}/secrets/{secret_id}/versions",
        )
        kwargs = {
            "headers": self.auth_headers,
            "params": {"pageSize": page_size, "pageToken": page_token},
            "response_model": SecretVersionsList,
            "raise_for_status": raise_for_status,
        }

        if iterator:
            return self._seekable_response(self.adapter.request, "versions", *args, **kwargs)

        response = self.adapter.request(*args, **kwargs)
        self._inject_client_to_items(response.versions)
        return response

    def schedule_secret_version_destruction(
        self, secret_id: str, version_id: str, pending_period: int = 604800, raise_for_status: bool = True
    ) -> Operation | YandexCloudError:
        """
        Schedules the specified version for destruction.
        Scheduled destruction can be cancelled with the :func:`cancel_secret_version_destruction()` method.

        :param secret_id: Secret indentifier.
        :param version_id: ID of the version to be destroyed.
        :param pending_period: Time interval in seconds between the version destruction request and actual destruction.
            Default value: ``604800`` (i.e. 7 days).
        :param raise_for_status: If set to ``False`` returns :class:`YandexCloudError` instead throw exception.
            Defaults to ``True``.
        """
        if isinstance(pending_period, int):
            if pending_period <= 0:
                raise ValueError("The ``pending_period`` value must be greater than 0.")
            # protobuf duration compat
            # https://github.com/protocolbuffers/protobuf/blob/main/src/google/protobuf/duration.proto
            pending_period = str(pending_period) + "s"
        else:
            raise ValueError("The ``pending_period`` value must be integer.")

        url = f"{self.lockbox_base_url}/secrets/{secret_id}:scheduleVersionDestruction"
        payload = {"versionId": version_id, "pendingPeriod": pending_period}
        response = self.adapter.request(
            "POST",
            url,
            headers=self.auth_headers,
            json=payload,
            response_model=Operation,
            raise_for_status=raise_for_status,
        )
        response.inject_client(self)
        return response

    # TODO: implement
    def set_secret_access_bindings(self, *args, **kwargs):
        """Not ready yet."""
        raise NotImplementedError

    def update_secret(
        self, secret_id: str, data: IUpdateSecret, raise_for_status: bool = True
    ) -> Operation | YandexCloudError:
        """
        Updates the specified secret.

        :param secret_id: Secret identifier.
        :param data: A new data for the secret as object.
            Important. Field mask that specifies which attributes of the secret are going to be updated.
            A comma-separated names off ALL fields to be updated. Only the specified fields will be changed.
            The others will be left untouched. If the field is specified in updateMask and no value for
            that field was sent in the request, the field's value will be reset to the default.
            The default value for most fields is null or 0.
            If ``updateMask`` is not sent in the request, all fields values will be updated.
            Fields specified in the request will be updated to provided values. The rest of the fields will be reset to the default.
        :param raise_for_status: If set to ``False`` returns :class:`YandexCloudError` instead throw exception.
            Defaults to ``True``.
        """
        url = f"{self.lockbox_base_url}/secrets/{secret_id}"
        payload = data.model_dump_json(by_alias=True)
        response = self.adapter.request(
            "PATCH",
            url,
            headers=self.auth_headers,
            data=payload,
            response_model=Operation,
            raise_for_status=raise_for_status,
        )
        response.inject_client(self)
        return response

    # TODO: implement
    def update_secret_access_bindings(self, *args, **kwargs):
        """Not ready yet."""
        raise NotImplementedError


# TODO: implement
class AsyncYandexLockboxClient(AbstractYandexLockboxClient):
    """The same as :class:`YandexLockboxClient` but async."""

    def __init__(self, *args, **kwargs) -> None:
        raise NotImplementedError  # pragma: no cover


# TODO: implement
class YandexLockbox:
    """
    A facade for encapsulating the logic of synchronous and asynchronous client operations,
    providing uniform methods.
    """

    def __init__(self) -> None:
        raise NotImplementedError  # pragma: no cover


__all__ = ["AsyncYandexLockboxClient", "YandexLockboxClient", "YandexLockbox"]



import logging
from typing import Any, Iterator, Union
from datetime import datetime
from pydantic import BaseModel, ConfigDict, Field, SecretStr, SecretBytes, computed_field

from yc_lockbox._constants import RpcError
from yc_lockbox._abc import AbstractYandexLockboxClient
from yc_lockbox._types import T
from yc_lockbox._exceptions import LockboxError


logger = logging.getLogger(__name__)


class BaseDomainModel(BaseModel):
    client: AbstractYandexLockboxClient | None = Field(
        None, description="Injected lockbox client for call model commands."
    )
    model_config: ConfigDict = ConfigDict(extra="ignore", arbitrary_types_allowed=True)  # type: ignore[union-attr]

    def inject_client(self, client: AbstractYandexLockboxClient) -> None:
        """
        Inject initialized client for make operations via domain model.

        :param inject_client: An initialized instance of :class:`AbstractYandexLockboxClient`.
        """
        self.client = client

    def _raise_when_empty_client(self) -> None:
        if self.client is None:  # pragma: no cover
            raise LockboxError("Lockbox client didn't injected to this resource.")


class BaseUpsertModel(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="forbid", populate_by_name=True)  # type: ignore[union-attr]


class IamTokenResponse(BaseDomainModel):
    token: str = Field(..., alias="iamToken")
    expires_at: datetime = Field(..., alias="expiresAt")


class SecretPayloadEntry(BaseDomainModel):
    """
    Domain object that represents an entry in the :class:`SecretPayload`.
    """

    key: str
    text_value: SecretStr | None = Field(None, alias="textValue")
    binary_value: SecretStr | SecretBytes | None = Field(None, alias="binaryValue")

    def reveal_text_value(self) -> str:
        """Reveal a text value."""
        if self.text_value is None:
            return None

        return self.text_value.get_secret_value()

    def reveal_binary_value(self) -> bytes:
        """Reveal a binary value."""
        if self.binary_value is None:
            return None

        return self.binary_value.get_secret_value()


class INewSecretPayloadEntry(BaseUpsertModel):
    key: str
    text_value: str | None = Field(None, alias="textValue")
    binary_value: str | None = Field(None, alias="binaryValue")


class INewSecret(BaseUpsertModel):
    folder_id: str = Field(..., alias="folderId")
    name: str | None = None
    description: str | None = None
    labels: dict[str, str] | None = {}
    kms_key_id: str | None = Field(None, alias="kmsKeyId")
    version_description: str | None = Field(None, alias="versionDescription")
    version_payload_entries: list[INewSecretPayloadEntry] = Field(..., alias="versionPayloadEntries")
    deletion_protection: bool = Field(False, alias="deletionProtection")


class IUpdateSecret(BaseUpsertModel):
    update_mask: str = Field(..., alias="updateMask", description="Comma-separated model field names to be updated.")
    name: str | None = None
    description: str | None = None
    labels: dict[str, str] | None = None
    deletion_protection: bool | None = Field(None, alias="deletionProtection")


class INewSecretVersion(BaseUpsertModel):
    description: str | None = None
    base_version_id: str | None = Field(None, alias="baseVersionId")
    payload_entries: list[INewSecretPayloadEntry] = Field(..., alias="payloadEntries")

    model_config: ConfigDict = ConfigDict(extra="forbid", populate_by_name=True)


class SecretPayload(BaseDomainModel):
    """
    Domain object that represents a payload for :class:`Secret`.
    """

    version_id: str = Field(..., alias="versionId")
    entries: list[SecretPayloadEntry]  # todo: dynamic object with entry-attributes instead list?

    def __getitem__(self, key: str | int) -> SecretPayloadEntry | None:
        """
        Get entry by key. Dictionary like. Also, list index available.

        :param key: Entry key or index.
        :raises KeyError: When key not exists in entries.
        :raises IndexError: When index out of range.
        """
        if isinstance(key, int):
            return self.entries[key]

        value: SecretPayloadEntry | None = self.get(key, default=None)

        if value is None:
            entries = ", ".join(map(lambda entry: entry.key, self.entries))
            raise KeyError(f"Entry with name {key} not exists. Available entries: {entries}")

        return value

    def get(self, key: str, default: Any = None) -> SecretPayloadEntry | None:
        """
        Get entry object from payload.

        :param key: Entry key (name).
        :param default: Default return value if key not exists.
        """
        return next(filter(lambda entry: entry.key == key, self.entries), default)


class SecretVersion(BaseDomainModel):
    """
    Domain object that represents a version from :class:`Secret`.
    This object contains methods for call version commands.

    """

    id: str
    status: str = "UNKNOWN"  # todo: enum
    description: str | None = None
    secret_id: str = Field(..., alias="secretId")
    created_at: datetime = Field(..., alias="createdAt")
    destroy_at: datetime | None = Field(None, alias="destroyAt")
    payload_entry_keys: list[str] | None = Field(None, alias="payloadEntryKeys")

    # TODO: implement ``is_current`` property if possible
    # there may be compatibility issues with the subquery
    # in different modes (synchronous, asynchronous)

    def cancel_version_destruction(self, **kwargs) -> Union["Operation", "YandexCloudError"]:
        """Shortcut for cancel destruction for this version."""
        self._raise_when_empty_client()
        return self.client.cancel_secret_version_destruction(self.secret_id, self.id, **kwargs)

    def payload(self, **kwargs) -> Union["SecretPayload", "YandexCloudError"]:
        """Get payload from the current secret.."""
        self._raise_when_empty_client()
        return self.client.get_secret_payload(self.secret_id, self.id, **kwargs)

    def schedule_version_destruction(
        self, pending_period: int = 604800, **kwargs
    ) -> Union["Operation", "YandexCloudError"]:
        """Shortcut for schedule descruction for this version."""
        self._raise_when_empty_client()
        return self.client.schedule_secret_version_destruction(self.secret_id, self.id, pending_period, **kwargs)


class Secret(BaseDomainModel):
    """
    A root domain model that represents Lockbox Secret.
    This model contains commands (methods) for secret manipulate.

    Usage::

        # basic commands
        secret.deactivate()
        secret.activate()
        secret.delete()

        # get payload from Secret
        secret_payload = secret.payload()
        print(secret_payload["my_entry"])  # by default secret values is masked like ******
        print(secret_payload["my_entry"].reveal_text_value())  # show real value

        # get all secret versions and destruct olds
        for version in secret.list_versions(iterator=True):
            print(version)
            if version.id != secret.current_version.id:
                version.schedule_version_destruction()

        # update a secret
        new_data = IUpdateSecret(
            update_mask="name,description",
            name="new-secret-name",
            description="My secret"
        )
        update_operation = secret.update(new_data)

        if update_operation.done:  # or use secret.refresh()
            print(update_operation.resource.name, update_operation.resource.description)

    """

    id: str
    status: str = "UNKNOWN"  # todo: enum
    name: str | None = None
    folder_id: str = Field(..., alias="folderId")
    created_at: datetime = Field(..., alias="createdAt")
    description: str | None = None
    kms_key_id: str | None = Field(None, alias="kmsKeyId")
    current_version: SecretVersion = Field(..., alias="currentVersion")
    deletion_protection: bool = Field(..., alias="deletionProtection")
    labels: dict[str, str] | None = None

    def activate(self, **kwargs) -> Union["Operation", "YandexCloudError"]:
        """Shortcut for activate the current secret."""
        self._raise_when_empty_client()
        return self.client.activate_secret(self.id, **kwargs)

    def add_version(self, version: INewSecretVersion, **kwargs) -> Union["Operation", "YandexCloudError"]:
        """Shortcut for add a new version to the current secret."""
        self._raise_when_empty_client()
        return self.client.add_secret_version(self.id, version, **kwargs)

    def cancel_version_destruction(self, version_id: str, **kwargs) -> Union["Operation", "YandexCloudError"]:
        """Shortcut for cancel destruction specified version of the current secret."""
        self._raise_when_empty_client()
        return self.client.cancel_secret_version_destruction(self.id, version_id, **kwargs)

    def deactivate(self, **kwargs) -> Union["Operation", "YandexCloudError"]:
        """Shortcut for deactivate the current secret."""
        self._raise_when_empty_client()
        return self.client.deactivate_secret(self.id, **kwargs)

    def delete(self, **kwargs) -> Union["Operation", "YandexCloudError"]:
        """Shortcut for delete the current secret."""
        self._raise_when_empty_client()
        return self.client.delete_secret(self.id, **kwargs)

    def refresh(self, **kwargs) -> "Secret":
        """Shortcut for refresh attributes for this secret."""
        self._raise_when_empty_client()
        data = self.client.get_secret(self.id, **kwargs)

        for attr, value in data.model_dump().items():
            if value != getattr(self, attr):
                setattr(self, attr, value)

        return data

    def payload(self, version_id: str | None = None, **kwargs) -> Union["Operation", "YandexCloudError"]:
        self._raise_when_empty_client()
        return self.client.get_secret_payload(self.id, version_id, **kwargs)

    def list_versions(
        self, page_size: int = 100, page_token: str | None = None, iterator: bool = False, **kwargs
    ) -> Union["SecretVersionsList", Iterator["SecretVersion"], "YandexCloudError"]:
        """Shortcut for list all available versions of the current secret."""
        self._raise_when_empty_client()
        return self.client.list_secret_versions(
            self.id, page_size=page_size, page_token=page_token, iterator=iterator, **kwargs
        )

    def schedule_version_destruction(
        self, version_id: str, pending_period: int = 604800, **kwargs
    ) -> Union["Operation", "YandexCloudError"]:
        """Shortcut for schedule destruction for specified version of the current secret."""
        self._raise_when_empty_client()
        return self.client.schedule_secret_version_destruction(self.id, version_id, pending_period, **kwargs)

    def update(self, data: IUpdateSecret, **kwargs) -> Union["Operation", "YandexCloudError"]:
        """Shortcut for update current secret."""
        self._raise_when_empty_client()
        return self.client.update_secret(self.id, data, **kwargs)


class BasePaginatedResponse(BaseDomainModel):
    next_page_token: str | None = Field(None, alias="nextPageToken")


class SecretsList(BasePaginatedResponse):
    secrets: list[Secret] = []


class SecretVersionsList(BasePaginatedResponse):
    versions: list[SecretVersion] = []


class YandexCloudError(BaseDomainModel):
    code: int
    message: str | None = None
    details: Any = None

    @computed_field
    @property
    def error_type(self) -> RpcError:
        try:
            value = RpcError(self.code)
        except ValueError:
            value = None

        return value


class Operation(BaseDomainModel):
    id: str
    done: bool = False
    description: str | None = None
    created_by: str | None = Field(None, alias="createdBy")
    created_at: datetime | None = Field(None, alias="createdAt")
    modified_at: datetime | None = Field(None, alias="modifiedAt")
    metadata: dict[str, str] | Any = None
    response: Any = None
    error: Any = None

    @computed_field
    @property
    def resource(self) -> T | None:
        """
        Returns response from operation if possible.
        Otherwise returns None.
        """
        if not self.done or not isinstance(self.response, dict):
            return None

        try:
            resource_type = self.response["@type"]
        except KeyError:
            return None

        match resource_type.split("/")[-1]:
            case "yandex.cloud.lockbox.v1.Secret":
                resource = Secret(**self.response)
            case "yandex.cloud.lockbox.v1.Version":
                resource = SecretVersion(**self.response)
            case "yandex.cloud.lockbox.v1.Payload":
                resource = SecretPayload(**self.response)
            case _:
                return None

        if self.client is not None:
            resource.inject_client(self.client)
        return resource



class LockboxError(Exception):
    def __init__(self, message: str, **kwargs) -> None:
        super().__init__(message, **kwargs)


class InvalidCredentials(LockboxError):
    def __init__(self, message: str = "Invalid credentials", **kwargs) -> None:
        super().__init__(message, **kwargs)


class BadCredentials(LockboxError):
    def __init__(self, message: str = "Bad credentials.", **kwargs) -> None:
        super().__init__(message, **kwargs)



import logging
from functools import cached_property
from datetime import datetime

from yc_lockbox._abc import AbstractYandexAuthClient, CredentialsType
from yc_lockbox._exceptions import BadCredentials
from yc_lockbox._adapters import HTTPAdapter
from yc_lockbox._models import IamTokenResponse

logger = logging.getLogger(__name__)


class YandexAuthClient(AbstractYandexAuthClient):
    """
    This is a simple client that allows you to get an up-to-date IAM token
    to make authenticated requests to Yandex Cloud.
    If you pass a IAM token as credentials, you need to take care
    of the freshness of the token yourself.

    :param credentials: Credentials for authenticate requests.
        Allowed types: service account key, OAuth token, IAM token.
    :param auth_base_url: Base IAM url without resource path URL.

    .. note::

        Important.
        This client works only in synchronous mode for backward compatibility.
    """

    def __init__(
        self,
        credentials: str | dict[str, str],
        *,
        auth_base_url: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(credentials, auth_base_url=auth_base_url, **kwargs)
        self.iam_expires_at = 0.0
        self.iam_token_url_path = self.auth_base_url + "/iam/v1/tokens"

    @cached_property
    def adapter(self) -> HTTPAdapter:
        """Returns HTTP adapter for communicate with Yandex Cloud."""
        return HTTPAdapter

    def get_iam_token(self) -> str:
        """Cacheable (in-memory, per instance) method for get IAM token from Yandex Cloud."""
        if self._credentials_type == CredentialsType.IAM_TOKEN:
            return self.credentials

        if self.iam_token is not None and self.iam_expires_at >= datetime.now().timestamp():
            # returns cached token
            return self.iam_token

        try:
            if self._credentials_type == CredentialsType.SERVICE_ACCOUNT:
                body = {"jwt": self._generate_jwt_for_sa()}
                response = self.adapter.request(
                    "POST", self.iam_token_url_path, json=body, response_model=IamTokenResponse
                )
                self.iam_token = response.token
                self.iam_expires_at = response.expires_at.timestamp()
                return self.iam_token

            elif self._credentials_type == CredentialsType.OAUTH_TOKEN:
                body = {"yandexPassportOauthToken": self.credentials}
                response = self.adapter.request(
                    "POST", self.iam_token_url_path, json=body, response_model=IamTokenResponse
                )
                self.iam_token = response.token
                self.iam_expires_at = response.expires_at.timestamp()
                return self.iam_token

        except Exception:
            raise BadCredentials("Bad credentials, can't issue IAM token.")
